@@ -3,7 +3,7 @@
  *			  procedural language
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/pl/plpgsql/src/pl_funcs.c,v 1.17.2.1 2002/05/05 17:38:28 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/pl/plpgsql/src/pl_funcs.c,v 1.38 2004/10/10 23:37:45 neilc Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -35,17 +35,13 @@
  *
  **********************************************************************/
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <unistd.h>
-#include <fcntl.h>
-#include <string.h>
-#include <ctype.h>
-
 #include "plpgsql.h"
 #include "pl.tab.h"
 
+#include <ctype.h>
+
+#include "parser/scansup.h"
+
 
 /* ----------
  * Local variables for the namestack handling
@@ -60,7 +56,7 @@ static bool ns_localmode = false;
  * ----------
  */
 void
-plpgsql_dstring_init(PLpgSQL_dstring * ds)
+plpgsql_dstring_init(PLpgSQL_dstring *ds)
 {
 	ds->value = palloc(ds->alloc = 512);
 	ds->used = 0;
@@ -73,7 +69,7 @@ plpgsql_dstring_init(PLpgSQL_dstring * ds)
  * ----------
  */
 void
-plpgsql_dstring_free(PLpgSQL_dstring * ds)
+plpgsql_dstring_free(PLpgSQL_dstring *ds)
 {
 	pfree(ds->value);
 }
@@ -84,7 +80,7 @@ plpgsql_dstring_free(PLpgSQL_dstring * ds)
  * ----------
  */
 void
-plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
+plpgsql_dstring_append(PLpgSQL_dstring *ds, const char *str)
 {
 	int			len = strlen(str);
 	int			needed = ds->used + len + 1;
@@ -92,7 +88,8 @@ plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
 	if (needed > ds->alloc)
 	{
 		/* might have to double more than once, if len is large */
-		do {
+		do
+		{
 			ds->alloc *= 2;
 		} while (needed > ds->alloc);
 		ds->value = repalloc(ds->value, ds->alloc);
@@ -108,7 +105,7 @@ plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
  * ----------
  */
 char *
-plpgsql_dstring_get(PLpgSQL_dstring * ds)
+plpgsql_dstring_get(PLpgSQL_dstring *ds)
 {
 	return ds->value;
 }
@@ -152,6 +149,9 @@ plpgsql_ns_push(char *label)
 {
 	PLpgSQL_ns *new;
 
+	if (label == NULL)
+		label = "";
+
 	new = palloc(sizeof(PLpgSQL_ns));
 	memset(new, 0, sizeof(PLpgSQL_ns));
 	new->upper = ns_current;
@@ -166,7 +166,7 @@ plpgsql_ns_push(char *label)
  * ----------
  */
 void
-plpgsql_ns_pop()
+plpgsql_ns_pop(void)
 {
 	int			i;
 	PLpgSQL_ns *old;
@@ -192,9 +192,7 @@ plpgsql_ns_additem(int itemtype, int itemno, char *name)
 	PLpgSQL_ns *ns = ns_current;
 	PLpgSQL_nsitem *nse;
 
-	if (name == NULL)
-		name = "";
-	name = plpgsql_tolower(name);
+	Assert(name != NULL);
 
 	if (ns->items_used == ns->items_alloc)
 	{
@@ -269,9 +267,7 @@ plpgsql_ns_lookup(char *name, char *label)
 				return ns->items[i];
 		}
 		if (ns_localmode)
-		{
 			return NULL;		/* name not found in current namespace */
-		}
 	}
 
 	return NULL;
@@ -317,51 +313,108 @@ plpgsql_ns_rename(char *oldname, char *newname)
 		}
 	}
 
-	elog(ERROR, "there is no variable '%s' in the current block", oldname);
+	ereport(ERROR,
+			(errcode(ERRCODE_UNDEFINED_OBJECT),
+			 errmsg("there is no variable \"%s\" in the current block",
+					oldname)));
 }
 
 
 /* ----------
- * plpgsql_tolower			Translate a string to lower case
- *					but honor "" escaping.
+ * plpgsql_convert_ident
+ *
+ * Convert a possibly-qualified identifier to internal form: handle
+ * double quotes, translate to lower case where not inside quotes,
+ * truncate to NAMEDATALEN.
+ *
+ * There may be several identifiers separated by dots and optional
+ * whitespace.	Each one is converted to a separate palloc'd string.
+ * The caller passes the expected number of identifiers, as well as
+ * a char* array to hold them.	It is an error if we find the wrong
+ * number of identifiers (cf grammar processing of fori_varname).
+ *
+ * NOTE: the input string has already been accepted by the flex lexer,
+ * so we don't need a heckuva lot of error checking here.
  * ----------
  */
-char *
-plpgsql_tolower(char *s)
+void
+plpgsql_convert_ident(const char *s, char **output, int numidents)
 {
-	char	   *sstart = s;
-	char	   *ret;
-	char	   *cp;
-
-	ret = palloc(strlen(s) + 1);
-	cp = ret;
+	const char *sstart = s;
+	int			identctr = 0;
 
+	/* Outer loop over identifiers */
 	while (*s)
 	{
+		char	   *curident;
+		char	   *cp;
+
+		/* Process current identifier */
+
 		if (*s == '"')
 		{
+			/* Quoted identifier: copy, collapsing out doubled quotes */
+
+			curident = palloc(strlen(s) + 1);	/* surely enough room */
+			cp = curident;
 			s++;
 			while (*s)
 			{
 				if (*s == '"')
-					break;
+				{
+					if (s[1] != '"')
+						break;
+					s++;
+				}
 				*cp++ = *s++;
 			}
-			if (*s != '"')
-				elog(ERROR, "unterminated \" in name %s", sstart);
+			if (*s != '"')		/* should not happen if lexer checked */
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("unterminated \" in name: %s", sstart)));
 			s++;
+			*cp = '\0';
+			/* Truncate to NAMEDATALEN */
+			truncate_identifier(curident, cp - curident, false);
 		}
 		else
 		{
-			if (isupper((unsigned char) *s))
-				*cp++ = tolower((unsigned char) *s++);
-			else
-				*cp++ = *s++;
+			/* Normal identifier: extends till dot or whitespace */
+			const char *thisstart = s;
+
+			while (*s && *s != '.' && !isspace((unsigned char) *s))
+				s++;
+			/* Downcase and truncate to NAMEDATALEN */
+			curident = downcase_truncate_identifier(thisstart, s - thisstart,
+													false);
+		}
+
+		/* Pass ident to caller */
+		if (identctr < numidents)
+			output[identctr++] = curident;
+		else
+			ereport(ERROR,
+					(errcode(ERRCODE_SYNTAX_ERROR),
+				   errmsg("qualified identifier cannot be used here: %s",
+						  sstart)));
+
+		/* If not done, skip whitespace, dot, whitespace */
+		if (*s)
+		{
+			while (*s && isspace((unsigned char) *s))
+				s++;
+			if (*s++ != '.')
+				elog(ERROR, "expected dot between identifiers: %s", sstart);
+			while (*s && isspace((unsigned char) *s))
+				s++;
+			if (*s == '\0')
+				elog(ERROR, "expected another identifier: %s", sstart);
 		}
 	}
-	*cp = '\0';
 
-	return ret;
+	if (identctr != numidents)
+		elog(ERROR, "improperly qualified identifier: %s",
+			 sstart);
 }
 
 
@@ -369,7 +422,7 @@ plpgsql_tolower(char *s)
  * Statement type as a string, for use in error messages etc.
  */
 const char *
-plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
+plpgsql_stmt_typename(PLpgSQL_stmt *stmt)
 {
 	switch (stmt->cmd_type)
 	{
@@ -393,6 +446,8 @@ plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
 			return "exit";
 		case PLPGSQL_STMT_RETURN:
 			return "return";
+		case PLPGSQL_STMT_RETURN_NEXT:
+			return "return next";
 		case PLPGSQL_STMT_RAISE:
 			return "raise";
 		case PLPGSQL_STMT_EXECSQL:
@@ -409,6 +464,8 @@ plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
 			return "fetch";
 		case PLPGSQL_STMT_CLOSE:
 			return "close";
+		case PLPGSQL_STMT_PERFORM:
+			return "perform";
 	}
 
 	return "unknown";
@@ -421,30 +478,32 @@ plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
 static int	dump_indent;
 
 static void dump_ind();
-static void dump_stmt(PLpgSQL_stmt * stmt);
-static void dump_block(PLpgSQL_stmt_block * block);
-static void dump_assign(PLpgSQL_stmt_assign * stmt);
-static void dump_if(PLpgSQL_stmt_if * stmt);
-static void dump_loop(PLpgSQL_stmt_loop * stmt);
-static void dump_while(PLpgSQL_stmt_while * stmt);
-static void dump_fori(PLpgSQL_stmt_fori * stmt);
-static void dump_fors(PLpgSQL_stmt_fors * stmt);
-static void dump_select(PLpgSQL_stmt_select * stmt);
-static void dump_exit(PLpgSQL_stmt_exit * stmt);
-static void dump_return(PLpgSQL_stmt_return * stmt);
-static void dump_raise(PLpgSQL_stmt_raise * stmt);
-static void dump_execsql(PLpgSQL_stmt_execsql * stmt);
-static void dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt);
-static void dump_dynfors(PLpgSQL_stmt_dynfors * stmt);
-static void dump_getdiag(PLpgSQL_stmt_getdiag * stmt);
-static void dump_open(PLpgSQL_stmt_open * stmt);
-static void dump_fetch(PLpgSQL_stmt_fetch * stmt);
-static void dump_close(PLpgSQL_stmt_close * stmt);
-static void dump_expr(PLpgSQL_expr * expr);
+static void dump_stmt(PLpgSQL_stmt *stmt);
+static void dump_block(PLpgSQL_stmt_block *block);
+static void dump_assign(PLpgSQL_stmt_assign *stmt);
+static void dump_if(PLpgSQL_stmt_if *stmt);
+static void dump_loop(PLpgSQL_stmt_loop *stmt);
+static void dump_while(PLpgSQL_stmt_while *stmt);
+static void dump_fori(PLpgSQL_stmt_fori *stmt);
+static void dump_fors(PLpgSQL_stmt_fors *stmt);
+static void dump_select(PLpgSQL_stmt_select *stmt);
+static void dump_exit(PLpgSQL_stmt_exit *stmt);
+static void dump_return(PLpgSQL_stmt_return *stmt);
+static void dump_return_next(PLpgSQL_stmt_return_next *stmt);
+static void dump_raise(PLpgSQL_stmt_raise *stmt);
+static void dump_execsql(PLpgSQL_stmt_execsql *stmt);
+static void dump_dynexecute(PLpgSQL_stmt_dynexecute *stmt);
+static void dump_dynfors(PLpgSQL_stmt_dynfors *stmt);
+static void dump_getdiag(PLpgSQL_stmt_getdiag *stmt);
+static void dump_open(PLpgSQL_stmt_open *stmt);
+static void dump_fetch(PLpgSQL_stmt_fetch *stmt);
+static void dump_close(PLpgSQL_stmt_close *stmt);
+static void dump_perform(PLpgSQL_stmt_perform *stmt);
+static void dump_expr(PLpgSQL_expr *expr);
 
 
 static void
-dump_ind()
+dump_ind(void)
 {
 	int			i;
 
@@ -453,7 +512,7 @@ dump_ind()
 }
 
 static void
-dump_stmt(PLpgSQL_stmt * stmt)
+dump_stmt(PLpgSQL_stmt *stmt)
 {
 	printf("%3d:", stmt->lineno);
 	switch (stmt->cmd_type)
@@ -488,6 +547,9 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_RETURN:
 			dump_return((PLpgSQL_stmt_return *) stmt);
 			break;
+		case PLPGSQL_STMT_RETURN_NEXT:
+			dump_return_next((PLpgSQL_stmt_return_next *) stmt);
+			break;
 		case PLPGSQL_STMT_RAISE:
 			dump_raise((PLpgSQL_stmt_raise *) stmt);
 			break;
@@ -512,14 +574,28 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_CLOSE:
 			dump_close((PLpgSQL_stmt_close *) stmt);
 			break;
+		case PLPGSQL_STMT_PERFORM:
+			dump_perform((PLpgSQL_stmt_perform *) stmt);
+			break;
 		default:
-			elog(ERROR, "plpgsql_dump: unknown cmd_type %d\n", stmt->cmd_type);
+			elog(ERROR, "unrecognized cmd_type: %d", stmt->cmd_type);
 			break;
 	}
 }
 
 static void
-dump_block(PLpgSQL_stmt_block * block)
+dump_stmts(PLpgSQL_stmts *stmts)
+{
+	int			i;
+
+	dump_indent += 2;
+	for (i = 0; i < stmts->stmts_used; i++)
+		dump_stmt(stmts->stmts[i]);
+	dump_indent -= 2;
+}
+
+static void
+dump_block(PLpgSQL_stmt_block *block)
 {
 	int			i;
 	char	   *name;
@@ -532,17 +608,34 @@ dump_block(PLpgSQL_stmt_block * block)
 	dump_ind();
 	printf("BLOCK <<%s>>\n", name);
 
-	dump_indent += 2;
-	for (i = 0; i < block->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (block->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(block->body);
+
+	if (block->exceptions)
+	{
+		for (i = 0; i < block->exceptions->exceptions_used; i++)
+		{
+			PLpgSQL_exception *exc = block->exceptions->exceptions[i];
+			PLpgSQL_condition *cond;
+
+			dump_ind();
+			printf("    EXCEPTION WHEN ");
+			for (cond = exc->conditions; cond; cond = cond->next)
+			{
+				if (cond != exc->conditions)
+					printf(" OR ");
+				printf("%s", cond->condname);
+			}
+			printf(" THEN\n");
+			dump_stmts(exc->action);
+		}
+	}
 
 	dump_ind();
 	printf("    END -- %s\n", name);
 }
 
 static void
-dump_assign(PLpgSQL_stmt_assign * stmt)
+dump_assign(PLpgSQL_stmt_assign *stmt)
 {
 	dump_ind();
 	printf("ASSIGN var %d := ", stmt->varno);
@@ -551,73 +644,53 @@ dump_assign(PLpgSQL_stmt_assign * stmt)
 }
 
 static void
-dump_if(PLpgSQL_stmt_if * stmt)
+dump_if(PLpgSQL_stmt_if *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("IF ");
 	dump_expr(stmt->cond);
 	printf(" THEN\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->true_body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->true_body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->true_body);
 
 	dump_ind();
 	printf("    ELSE\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->false_body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->false_body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->false_body);
 
 	dump_ind();
 	printf("    ENDIF\n");
 }
 
 static void
-dump_loop(PLpgSQL_stmt_loop * stmt)
+dump_loop(PLpgSQL_stmt_loop *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("LOOP\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDLOOP\n");
 }
 
 static void
-dump_while(PLpgSQL_stmt_while * stmt)
+dump_while(PLpgSQL_stmt_while *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("WHILE ");
 	dump_expr(stmt->cond);
 	printf("\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDWHILE\n");
 }
 
 static void
-dump_fori(PLpgSQL_stmt_fori * stmt)
+dump_fori(PLpgSQL_stmt_fori *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("FORI %s %s\n", stmt->var->refname, (stmt->reverse) ? "REVERSE" : "NORMAL");
 
@@ -630,36 +703,30 @@ dump_fori(PLpgSQL_stmt_fori * stmt)
 	printf("    upper = ");
 	dump_expr(stmt->upper);
 	printf("\n");
-
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
 	dump_indent -= 2;
 
+	dump_stmts(stmt->body);
+
 	dump_ind();
 	printf("    ENDFORI\n");
 }
 
 static void
-dump_fors(PLpgSQL_stmt_fors * stmt)
+dump_fors(PLpgSQL_stmt_fors *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("FORS %s ", (stmt->rec != NULL) ? stmt->rec->refname : stmt->row->refname);
 	dump_expr(stmt->query);
 	printf("\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDFORS\n");
 }
 
 static void
-dump_select(PLpgSQL_stmt_select * stmt)
+dump_select(PLpgSQL_stmt_select *stmt)
 {
 	dump_ind();
 	printf("SELECT ");
@@ -682,7 +749,7 @@ dump_select(PLpgSQL_stmt_select * stmt)
 }
 
 static void
-dump_open(PLpgSQL_stmt_open * stmt)
+dump_open(PLpgSQL_stmt_open *stmt)
 {
 	dump_ind();
 	printf("OPEN curvar=%d\n", stmt->curvar);
@@ -714,7 +781,7 @@ dump_open(PLpgSQL_stmt_open * stmt)
 }
 
 static void
-dump_fetch(PLpgSQL_stmt_fetch * stmt)
+dump_fetch(PLpgSQL_stmt_fetch *stmt)
 {
 	dump_ind();
 	printf("FETCH curvar=%d\n", stmt->curvar);
@@ -735,14 +802,23 @@ dump_fetch(PLpgSQL_stmt_fetch * stmt)
 }
 
 static void
-dump_close(PLpgSQL_stmt_close * stmt)
+dump_close(PLpgSQL_stmt_close *stmt)
 {
 	dump_ind();
 	printf("CLOSE curvar=%d\n", stmt->curvar);
 }
 
 static void
-dump_exit(PLpgSQL_stmt_exit * stmt)
+dump_perform(PLpgSQL_stmt_perform *stmt)
+{
+	dump_ind();
+	printf("PERFORM expr = ");
+	dump_expr(stmt->expr);
+	printf("\n");
+}
+
+static void
+dump_exit(PLpgSQL_stmt_exit *stmt)
 {
 	dump_ind();
 	printf("EXIT lbl='%s'", stmt->label);
@@ -755,24 +831,37 @@ dump_exit(PLpgSQL_stmt_exit * stmt)
 }
 
 static void
-dump_return(PLpgSQL_stmt_return * stmt)
+dump_return(PLpgSQL_stmt_return *stmt)
 {
 	dump_ind();
 	printf("RETURN ");
-	if (stmt->retrecno > 0)
+	if (stmt->retrecno >= 0)
 		printf("record %d", stmt->retrecno);
+	else if (stmt->retrowno >= 0)
+		printf("row %d", stmt->retrowno);
+	else if (stmt->expr == NULL)
+		printf("NULL");
 	else
-	{
-		if (stmt->expr == NULL)
-			printf("NULL");
-		else
-			dump_expr(stmt->expr);
-	}
+		dump_expr(stmt->expr);
 	printf("\n");
 }
 
 static void
-dump_raise(PLpgSQL_stmt_raise * stmt)
+dump_return_next(PLpgSQL_stmt_return_next *stmt)
+{
+	dump_ind();
+	printf("RETURN NEXT ");
+	if (stmt->rec != NULL)
+		printf("target = %d %s\n", stmt->rec->recno, stmt->rec->refname);
+	else if (stmt->row != NULL)
+		printf("target = %d %s\n", stmt->row->rowno, stmt->row->refname);
+	else if (stmt->expr != NULL)
+		dump_expr(stmt->expr);
+	printf("\n");
+}
+
+static void
+dump_raise(PLpgSQL_stmt_raise *stmt)
 {
 	int			i;
 
@@ -784,7 +873,7 @@ dump_raise(PLpgSQL_stmt_raise * stmt)
 }
 
 static void
-dump_execsql(PLpgSQL_stmt_execsql * stmt)
+dump_execsql(PLpgSQL_stmt_execsql *stmt)
 {
 	dump_ind();
 	printf("EXECSQL ");
@@ -793,7 +882,7 @@ dump_execsql(PLpgSQL_stmt_execsql * stmt)
 }
 
 static void
-dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt)
+dump_dynexecute(PLpgSQL_stmt_dynexecute *stmt)
 {
 	dump_ind();
 	printf("EXECUTE ");
@@ -802,26 +891,21 @@ dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt)
 }
 
 static void
-dump_dynfors(PLpgSQL_stmt_dynfors * stmt)
+dump_dynfors(PLpgSQL_stmt_dynfors *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("FORS %s EXECUTE ", (stmt->rec != NULL) ? stmt->rec->refname : stmt->row->refname);
 	dump_expr(stmt->query);
 	printf("\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDFORS\n");
 }
 
 static void
-dump_getdiag(PLpgSQL_stmt_getdiag * stmt)
+dump_getdiag(PLpgSQL_stmt_getdiag *stmt)
 {
 	int			i;
 
@@ -855,7 +939,7 @@ dump_getdiag(PLpgSQL_stmt_getdiag * stmt)
 }
 
 static void
-dump_expr(PLpgSQL_expr * expr)
+dump_expr(PLpgSQL_expr *expr)
 {
 	int			i;
 
@@ -875,7 +959,7 @@ dump_expr(PLpgSQL_expr * expr)
 }
 
 void
-plpgsql_dumptree(PLpgSQL_function * func)
+plpgsql_dumptree(PLpgSQL_function *func)
 {
 	int			i;
 	PLpgSQL_datum *d;
@@ -928,8 +1012,9 @@ plpgsql_dumptree(PLpgSQL_function * func)
 					printf("ROW %-16s fields", row->refname);
 					for (i = 0; i < row->nfields; i++)
 					{
-						printf(" %s=var %d", row->fieldnames[i],
-							   row->varnos[i]);
+						if (row->fieldnames[i])
+							printf(" %s=var %d", row->fieldnames[i],
+								   row->varnos[i]);
 					}
 					printf("\n");
 				}
@@ -938,7 +1023,15 @@ plpgsql_dumptree(PLpgSQL_function * func)
 				printf("REC %s\n", ((PLpgSQL_rec *) d)->refname);
 				break;
 			case PLPGSQL_DTYPE_RECFIELD:
-				printf("RECFIELD %-16s of REC %d\n", ((PLpgSQL_recfield *) d)->fieldname, ((PLpgSQL_recfield *) d)->recno);
+				printf("RECFIELD %-16s of REC %d\n",
+					   ((PLpgSQL_recfield *) d)->fieldname,
+					   ((PLpgSQL_recfield *) d)->recparentno);
+				break;
+			case PLPGSQL_DTYPE_ARRAYELEM:
+				printf("ARRAYELEM of VAR %d subscript ",
+					   ((PLpgSQL_arrayelem *) d)->arrayparentno);
+				dump_expr(((PLpgSQL_arrayelem *) d)->subscript);
+				printf("\n");
 				break;
 			case PLPGSQL_DTYPE_TRIGARG:
 				printf("TRIGARG ");
@@ -955,4 +1048,5 @@ plpgsql_dumptree(PLpgSQL_function * func)
 	printf("%3d:", func->action->lineno);
 	dump_block(func->action);
 	printf("\nEnd of execution tree of function %s\n\n", func->fn_name);
+	fflush(stdout);
 }
